#! /usr/bin/python3
import subprocess
import sys
import time
import os

from sysrepo_upgrade import sysrepo_upgrade as su
from systemd import journal

au_stream = journal.stream(identifier='auto-upgrade')
nfvt_up = '/opt/bin/nfvt-up'
INITIAL_ZTP_FILE = "/config/_ztp/initial_ztp"


def check_config():
    params = su.SysrepoUpgrade(app_name="auto-upgrade")
    if not params.get_enabled():
        print('disabled. DO Nothing !', file=au_stream)
        sys.exit(0)

    target_version = params.get_target_version()
    if not target_version:
        print('WARNING missing target-version. Will try again in a while ...', file=au_stream)
        sys.exit(1)

    try:
        current_ver = subprocess.check_output(['dpkg-query', "--showformat='${Version}'", '--show', 'nfvtimeos'],
                       stderr=subprocess.DEVNULL)
    except subprocess.CalledProcessError as cpe:
        print('ERROR {}'.format(cpe), file=au_stream)
        sys.exit(0)
    except OSError:
        print('ERROR missing {}'.format(nfvt_up), file=au_stream)
        sys.exit(0)
    except Exception as e:
        print('ERROR {}'.format(e), file=au_stream)
        sys.exit(1)

    if target_version in str(current_ver):
        print('Already on version {}!'.format(target_version), file=au_stream)
        sys.exit(0)

    if not params.get_repo_servers():
        print('WARNING missing repo-server. Will try again in a while ...', file=au_stream)
        sys.exit(1)

    available_versions = '/var/log/available_versions.log'
    try:
        subprocess.run([nfvt_up, '-l', '-f', available_versions],
                       stderr=subprocess.DEVNULL)
    except subprocess.CalledProcessError as cpe:
        print('ERROR {}'.format(cpe), file=au_stream)
        sys.exit(0)
    except OSError:
        print('ERROR missing {}'.format(nfvt_up), file=au_stream)
        sys.exit(0)
    except Exception as e:
        print('ERROR {}'.format(e), file=au_stream)
        sys.exit(1)

    with open(available_versions, 'r') as f:
        reported_version_list = [myl for myl in (line.strip().replace(',', '').split() for line in f) if myl]
        reported_versions = reported_version_list[0][2:]

    if not reported_versions \
       or target_version not in reported_versions:
        print('WARNING version {} not in {}\nWill try again in a while ...\n'.format(
            target_version, reported_versions), file=au_stream)
        sys.exit(1)

    return target_version


def download(version):
    download_version = '/var/log/download_version.log'
    try:
        downl = subprocess.run([nfvt_up, '-v', version,
                                '--download', '-f', download_version],
                               stdout=subprocess.PIPE, encoding='UTF-8')
    except subprocess.CalledProcessError as cpe:
        print('ERROR {}'.format(cpe), file=au_stream)
        sys.exit(1)
    except OSError:
        print('missing {}'.format(nfvt_up), file=au_stream)
        sys.exit(1)

    if 0 != downl.returncode:
        with open(download_version, 'r') as f:
            lines = f.readlines()
            print(lines, file=au_stream)

        sys.exit(1)


def upgrade(version):
    upgrade_version = '/var/log/upgrade_version.log'
    try:
        upgr = subprocess.run([nfvt_up, '-v', version,
                               '-u', '-f', upgrade_version,
                               '--user', 'admin'],
                              stdout=subprocess.PIPE, encoding='UTF-8')
    except subprocess.CalledProcessError as cpe:
        print('ERROR {}'.format(cpe), file=au_stream)
        sys.exit(1)
    except OSError:
        print('missing {}'.format(nfvt_up), file=au_stream)
        sys.exit(1)

    if 0 != upgr.returncode:
        with open(upgrade_version, 'r') as f:
            lines = f.readlines()
            print(lines, file=au_stream)

        print('WARNING upgrade return code {}. Do nothing !!!'.format(upgr.returncode))
        sys.exit(0)


def systemd_verify_running():
    try:
        sd = subprocess.run(['/bin/systemctl', 'is-system-running'],
                            stdout=subprocess.PIPE, encoding='UTF-8')
    except subprocess.CalledProcessError as cpe:
        print('ERROR {}'.format(cpe), file=au_stream)
        sys.exit(1)
    except OSError:
        print('missing {}'.format(nfvt_up), file=au_stream)
        sys.exit(1)

    if 0 != sd.returncode:
        print('system is not in running state. Will try again in a while ...', file=au_stream)
        sys.exit(1)

    return


def systemd_no_restarting_services():
    try:
        sd = subprocess.run(['/bin/systemctl', 'list-jobs'],
                            stdout=subprocess.PIPE, encoding='UTF-8')
    except subprocess.CalledProcessError as cpe:
        print('ERROR {}'.format(cpe), file=au_stream)
        sys.exit(1)
    except OSError:
        print('missing /bin/systemctl', file=au_stream)
        sys.exit(1)

    if 0 != sd.returncode:
        print('WARNING systemctl list-jobs returns {}. Will try again in a while ...'.format(
            sd.returncode), file=au_stream)
        sys.exit(1)

    if 0 > sd.stdout.find('No jobs running'):
        print('WARNING system working jobs\n{}\nWill try again in a while ...'.format(
            sd.stdout), file=au_stream)
        sys.exit(1)


def reboot():
    print('WARNING Upgrade finished. Calling REBOOT !', file=au_stream)

    if not os.path.exists(INITIAL_ZTP_FILE):
        f = open(INITIAL_ZTP_FILE,"w+")
        f.close()

    if os.path.exists('/config/_reboot/tve/now'):
        os.rmdir('/config/_reboot/tve/now')

	# Create dirs
    os.makedirs('/config/_reboot/tve/now')


def main(_name, _argv):
    systemd_verify_running()
    systemd_no_restarting_services()

    target_version = check_config()

    download(version=target_version)
    upgrade(version=target_version)
    reboot()


if __name__ == "__main__":
    main(sys.argv[0], sys.argv[1:])
