#!/usr/bin/env python3
import json
import os
import re
import subprocess

import aptsources.sourceslist
from sysrepo_api import sysrepo_api3 as api
from sysrepo_system import sysrepo_system as srs

UCPE_STATUS_FILE = "/config/_hb/ucpe_status.json"
INITIAL_ZTP_FILE = "/config/_ztp/initial_ztp"

class SysrepoUpgrade(api.SysrepoAPI):
    CONST_DISTR = 'nfvtimeos'
    CONST_MODS = ['main']

    def __init__(self,
                 app_name="upgrade",
                 log_file=None,
                 log_level=None,
                 console_log=False,
                 indexes={"repo-server": "url"},
                 namespace="/ucpe:config/upgrade:upgrade",
                 run_as_subscriber=False,
                 priority=0):
        api.SysrepoAPI.__init__(
            self,
            app_name=app_name,
            log_file=log_file,
            log_level=log_level,
            console_log=console_log,
            namespace=namespace,
            indexes=indexes,
            run_as_subscriber=run_as_subscriber,
            priority=priority)
        self.enabled_change = None
        self.window_change = None

    def _remove_apt_source(self, url):
        self.logger.debug('_remove_apt_source(%s)' % url)
        sr = aptsources.sourceslist.SourcesList()

        to_remove = []
        for sl in sr:
            #if sl.disabled or sl.invalid:
            #    continue

            #if sl.dist == self.CONST_DISTR and url == sl.uri:
            #    to_remove.append(sl)
            to_remove.append(sl)

        if len(to_remove) > 0:
            for rem in to_remove:
                sr.remove(rem)

            sr.save()

    def _clean_apt_sources(self):
        self.logger.debug('_clean_apt_sources')
        sr = aptsources.sourceslist.SourcesList()

        to_remove = []

        for sl in sr:
            if sl.disabled or sl.invalid:
                continue

            if sl.dist == self.CONST_DISTR:
                to_remove.append(sl)

        for rem in to_remove:
            sr.remove(rem)

        sr.save()

    def _add_apt_source(self, url):
        self.logger.debug('_add_apt_source( %s )' % url)
        sr = aptsources.sourceslist.SourcesList()
        sr.add('deb', url, self.CONST_DISTR, self.CONST_MODS)
        sr.save()

    def auto_upgrade_service_ctl(self, enabled=False):
        operation = 'start' if enabled and 'true' in str(enabled).lower() else 'stop'
        self.logger.debug('auto_upgrade_service_control {}'.format(operation))
        try:
            subprocess.run(['/bin/systemctl', 'daemon-reload'],
                           stderr=subprocess.DEVNULL)
            subprocess.run(['/bin/systemctl', 'reset-failed',
                            'auto-upgrade.service'],
                           stderr=subprocess.DEVNULL)
            subprocess.run(['/bin/systemctl', '--no-block',
                            operation,
                            'auto-upgrade.service'],
                           stderr=subprocess.DEVNULL)
        except subprocess.CalledProcessError as cpe:
            self.logger.error('ERROR {}'.format(cpe))
        except OSError:
            self.logger.error('missing /bin/systemctl')

    def auto_upgrade_timer_ctl(self, enabled=False):
        operation = 'start' if enabled and 'true' in str(enabled).lower() else 'stop'
        boot = 'enable' if enabled and 'true' in str(enabled).lower() else 'disable'
        self.logger.debug('auto_upgrade_timer_control {} {}'.format(operation, boot))
        try:
            subprocess.run(['/bin/systemctl', 'daemon-reload'],
                           stderr=subprocess.DEVNULL)
            subprocess.run(['/bin/systemctl', 'reset-failed',
                            'auto-upgrade.service'],
                           stderr=subprocess.DEVNULL)
            subprocess.run(['/bin/systemctl', 'stop',
                            '--now', 'auto-upgrade.timer'],
                           stderr=subprocess.DEVNULL)
            subprocess.run(['/bin/systemctl', 'disable',
                            '--now', 'auto-upgrade.timer'],
                           stderr=subprocess.DEVNULL)
            subprocess.run(['/bin/systemctl', operation,
                            '--now', 'auto-upgrade.timer'],
                           stderr=subprocess.DEVNULL)
            subprocess.run(['/bin/systemctl', boot,
                            '--now', 'auto-upgrade.timer'],
                           stderr=subprocess.DEVNULL)
        except subprocess.CalledProcessError as cpe:
            self.logger.error('ERROR {}'.format(cpe))
        except OSError:
            self.logger.error('missing /bin/systemctl')

    def auto_upgrade_timer_set(self, new_time=''):
        self.logger.debug('timer_set {}'.format(new_time))
        try:
            subprocess.run(['/bin/sed', '-i',
                            's#^OnCalendar=.*$#OnCalendar={}#g'.format(new_time),
                            '/lib/systemd/system/auto-upgrade.timer'],
                           stderr=subprocess.DEVNULL)
        except subprocess.CalledProcessError as cpe:
            self.logger.error('ERROR {}'.format(cpe))
        except OSError:
            self.logger.error('missing /bin/systemctl')
        else:
            self.auto_upgrade_timer_ctl(enabled=True)

    @staticmethod
    def open_ippass(value=None, add_server=1):
        ippass = srs.SysrepoSystem(app_name="sr-upgrade_ippass")
        if value:
            ippass._SysrepoSystem__config_vpp_ippass('tcp', value, 80, add_server)

    def apply_startup(self):
        self.logger.debug('Apply STARTUP config for system')
        sys_conf = self.get_startup_config(full_config=False, full_path=False)

        self.logger.debug('Startup config : %s' % json.dumps(sys_conf))

        self._clean_apt_sources()

        if 'repo-server' in sys_conf and len(sys_conf['repo-server']) > 0:
            self.logger.debug('Startup URL repo-server server configured')
            for server in sys_conf['repo-server']:
                self._add_apt_source(server['url'])
                SysrepoUpgrade.open_ippass(server)  # vpp

        if 'enabled' in sys_conf:
            enabled = sys_conf['enabled']
            self.enabled_change = sys_conf['enabled']
            self.logger.debug('Startup enabled {}'.format(enabled))
            if enabled and 'true' in str(enabled).lower():
                # NFV-12784 - ztp completed + enabled
                if os.path.exists(UCPE_STATUS_FILE):
                    try:
                        with open(UCPE_STATUS_FILE, 'r') as f:
                            ucpe_ztp_status = json.loads(f.read())
                    except OSError as e:
                        self.logger.debug('Error {}'.format(e))
                    else:
                        self.logger.debug('ztp status {}'.format(ucpe_ztp_status['ucpe:status']['ztp-status']))
                        if 'ztp-status' in ucpe_ztp_status['ucpe:status'] \
                           and ucpe_ztp_status['ucpe:status']['ztp-status'] == 'completed':
                            # check system is 'starting' ?@
                            self.logger.debug('NFV-12784 ztp completed + enabled on startup')
                            if not os.path.exists(INITIAL_ZTP_FILE):
                                self.auto_upgrade_service_ctl(enabled=True)
                        else:
                            self.logger.debug('WARNING startup enabled but ztp unknown. DO Nothing !')
                            try:
                                ctl = subprocess.run(['/bin/systemctl', 'is-system-running'],
                                                     stdout=subprocess.PIPE, encoding='UTF-8')
                            except subprocess.CalledProcessError as cpe:
                                self.logger.debug('ERROR {}'.format(cpe))
                                pass
                            except OSError:
                                self.logger.debug('Missing /bin/systemctl')
                                pass

                            if 0 == ctl.returncode:
                                self.logger.debug('{}'.format(ctl.stdout))
                            else:
                                self.logger.debug('systemctl is-system-running return {}'.format(
                                    ctl.returncode))

            elif enabled == 'false' or not enabled:
                self.logger.debug('Startup Disable/Stop timer and service')
                self.auto_upgrade_timer_ctl(enabled=False)
                self.auto_upgrade_service_ctl(enabled=False)

        if 'target-version' in sys_conf:
            self.logger.debug('Startup target-version {}'.format(
                sys_conf['target-version']))
            # TODO
            pass

        if 'time-window' in sys_conf:
            self.logger.debug('Startup time-window {}'.format(
                sys_conf['time-window']))
            self.window_change = sys_conf['time-window']
            pass

        return True

    # Post apply actions callback
    def post_apply_config(self, change_set):
        self.logger.debug('Post RUNNING config')

        if not change_set:
            self.logger.debug('Post apply empty change_set, return')
            return

        if self.enabled_change is None:
            enabled = self.get_time_window()
        else:
            enabled = self.enabled_change
        self.enabled_change = None

        if self.window_change is None:
            time_window = self.get_time_window()
        else:
            time_window = self.window_change
        self.window_change = None

        self.logger.debug('Post enabled {} time_window {}'.format(enabled, time_window))
        if enabled and 'true' in str(enabled).lower():
            if 'anytime' in time_window.lower():
                # NFV-12785 enabled + 'anytime' == start now
                self.logger.debug('NFV-12785 apply + enabled + anytime')
                self.auto_upgrade_service_ctl(enabled=True)
            elif not os.path.exists(INITIAL_ZTP_FILE):
                # initial ZTP
                self.logger.debug('NFV-12785 apply + enabled + First ZTP ever')
                self.auto_upgrade_service_ctl(enabled=True)
            else:
                # NFV-12783
                time_of_day = re.match(r'^\d{1,2}:\d{1,2}$', time_window, re.MULTILINE | re.IGNORECASE)
                time_of_week_and_day = re.match(r'^(Monday|Tuesday|Wednesday|Thursday|Friday|Saturday|Sunday) \d{1,2}:\d{1,2}$',
                                                time_window, re.MULTILINE | re.IGNORECASE)
                date_and_time = re.match(r'^((Monday|Tuesday|Wednesday|Thursday|Friday|Saturday|Sunday) )?\d{4}-\d{1,2}-\d{1,2} \d{1,2}:\d{1,2}$',
                                         time_window, re.MULTILINE | re.IGNORECASE)
                if time_of_day:
                    self.logger.debug('Match time_of_day {}'.format(
                        time_of_day.group()))
                    self.auto_upgrade_timer_set(new_time=time_of_day.group())
                elif time_of_week_and_day:
                    self.logger.debug('Match time_of_week_and_day {}'.format(
                        time_of_week_and_day.group()))
                    self.auto_upgrade_timer_set(new_time=time_of_week_and_day.group())
                elif date_and_time:
                    self.logger.debug('Match date_and_time {}'.format(
                        date_and_time.group()))
                    self.auto_upgrade_timer_set(new_time=date_and_time.group())
                else:
                    self.logger.critical('ERROR unknown time_window format {}'.format(
                        time_window))

        elif enabled == 'false' or not enabled:
            self.logger.debug('Running Disable/Stop timer and service')
            self.auto_upgrade_timer_ctl(enabled=False)
            self.auto_upgrade_service_ctl(enabled=False)

        return

    def apply_running(self, change_set):
        self.logger.debug('RUNNING config')

        for change in change_set:
            if not change["xpath"].startswith(self.namespace):
                continue

            if 'operation' not in change:
                self.logger.error('Malformed change set')
                continue

            self.logger.debug('Operation -->> %s' % change['operation'])
            self.logger.debug('XPath -->> %s' % change['xpath'])

            if change['operation'] == 'CREATED':
                self.logger.debug('Created - %s' % change['xpath'])
                if change["xpath"].find("repo-server") != -1 and change["xpath"].find("/url") != -1:
                    self._add_apt_source(change['new_value'])
                    SysrepoUpgrade.open_ippass(change['new_value'])  # vpp

            elif change['operation'] == 'DELETED':
                self.logger.debug('Deleted - %s' % change['xpath'])
                if change["xpath"].find("repo-server") != -1 and change["xpath"].find("/url") != -1:
                    self.logger.debug('Deleted - %s, \'%s\'' % (change['xpath'], change['old_value']))
                    self._remove_apt_source(change['old_value'])
                    SysrepoUpgrade.open_ippass(change['old_value'], add_server=0)  # vpp

            elif change['operation'] == 'MODIFIED':
                self.logger.debug('Running Modified - {}: {} -> {}'.
                                  format(change['xpath'],
                                         change['old_value'],
                                         change['new_value']))
                if 'enabled' in change['xpath']:
                    enabled = change['new_value']
                    self.enabled_change = change['new_value']
                    self.logger.debug('Running enabled {}'.format(enabled))
                    if enabled == 'false' or not enabled:
                        self.auto_upgrade_timer_ctl(enabled=False)

                if 'target-version' in change["xpath"]:
                    self.logger.debug('Running target-version')
                    # TODO
                    pass

                if 'time-window' in change["xpath"]:
                    self.logger.debug('Running time-window')
                    self.window_change = change['new_value']
                    pass

            elif change['operation'] == 'MOVED':
                self.logger.debug('Moved  - %s' % change['xpath'])
            else:
                self.logger.error('Unknown operation \'%s\'' % change['operation'])

        return True

    def apply_config(self, change_set=None, apply_startup=False):
        if apply_startup:
            return self.apply_startup()
        else:
            return self.apply_running(change_set)

    def set_enabled(self, value):
        self.set_item('{}/enabled'.format(self.namespace), value)

    def get_enabled(self):
        return self.get_item('{}/enabled'.format(self.namespace))

    def unset_enabled(self):
        self.delete_item('{}/enabled'.format(self.namespace))

    def set_target_version(self, value):
        self.set_item('{}/target-version'.format(self.namespace), value)

    def get_target_version(self):
        return self.get_item('{}/target-version'.format(self.namespace))

    def unset_target_version(self):
        self.delete_item('{}/target-version'.format(self.namespace))

    def set_time_window(self, value):
        self.set_item('{}/time-window'.format(self.namespace), value)

    def get_time_window(self):
        return self.get_item('{}/time-window'.format(self.namespace))

    def unset_time_window(self):
        self.delete_item('{}/time-window'.format(self.namespace))

    def _remove_repo_servers(self):
        self._clean_apt_sources()
        repo_serv = self.get_config_values(xpath="{}/repo-server".format(self.namespace))
        if repo_serv is None:
            return

        for url in repo_serv:
            self.unset_repo_server(url)

    def set_repo_server(self, url):
        # We have list with one element and need to remove it
        self.set_item("{}/repo-server[url='{}']".format(self.namespace, url), url)

    def get_repo_server(self, url):
        return self.get_item("{}/repo-server[url='{}']".format(self.namespace, url))

    def unset_repo_server(self, url):
        self.delete_item("{}/repo-server[url='{}']".format(self.namespace, url))

    def unset_repo_all(self):
        self._remove_repo_servers()

    def get_repo_servers(self):
        config = self.get_running_config(full_path=False)
        config_indexed = {}
        if 'repo-server' not in config:
            return config_indexed
        for server in config['repo-server']:
            config_indexed[server['url']] = server
        return config_indexed

